using System;
using System.Xml;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace MemGraph
{
    public partial class MemGraph : Form
    {
        int fsizeX = 0;
        int fsizeY = 0;
        int x = 350;
        int xV = 350;
        int y = 30;
        int yV = 30;
        int dragStart = 0;
        int xStart = 0;
        int dragStartY = 0;
        int yStart = 0;
        bool dragging = false;
        ulong size;
        ulong sizeV;
        FormWindowState OldState;
        MemItem edit = null;
        private Font _font;

        public MemGraph()
        {
            InitializeComponent();
        }

        private void DelPhys_Click(object sender, EventArgs e)
        {
            if (LstPhys.SelectedItem != null)
                LstPhys.Items.Remove(LstPhys.SelectedItem);
            Redraw();
        }

        private void Del_Click(object sender, EventArgs e)
        {
            ListBox l = (MemTabs.SelectedIndex == 0) ? LstPhys : LstVirtual;
            if (l.SelectedItem != null)
                l.Items.Remove(l.SelectedItem);
            Redraw();
        }
                
        private void BtnAdd_Click(object sender, EventArgs e)
        {
            ListBox l = (MemTabs.SelectedIndex == 0) ? LstPhys : LstVirtual;
            MemItem itm = new MemItem();
            try
            {
                itm.Name = TxtName.Text;
                itm.From = Parse(TxtFrom.Text);
                itm.To = Parse(TxtTo.Text);
                itm.Color = LColor.BackColor;
            }
            catch
            {
                MessageBox.Show("Invalid values");
                return;
            }

            foreach (MemItem it in l.Items)
            {
                if ((itm.From < it.From && itm.To > it.From)
                    || (itm.From < it.To && itm.To > it.To)
                    || (itm.From >= it.From && itm.To <= it.To))
                {
                    MessageBox.Show("Collitions With: " + it.Name);
                }
            }

            InsertSorted(l, itm);

            TxtName.Text = "";
            TxtFrom.Text = "";
            TxtTo.Text = "";
            Redraw();
        }

        private void LColor_Click(object sender, EventArgs e)
        {
            if (DlgColor.ShowDialog() == DialogResult.OK)
            {
                LColor.BackColor = Color.FromArgb(190, DlgColor.Color.R, DlgColor.Color.G, DlgColor.Color.B);
            }
        }
                
        private void BtnSet_Click(object sender, EventArgs e)
        {
            if(MemTabs.SelectedIndex == 0)
                size = Parse(TxtSize.Text);
            else
                sizeV = Parse(TxtSizeV.Text);
            Redraw();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            PBox.Image = new Bitmap(PBox.Width, PBox.Height);
            PBoxV.Image = new Bitmap(PBoxV.Width, PBoxV.Height); 
            size = Parse(TxtSize.Text);
            sizeV = Parse(TxtSizeV.Text);
            _font = new Font(Font.FontFamily, 10, FontStyle.Regular, GraphicsUnit.Pixel);
            SizeUD.Value = (int)_font.Size;
            OldState = WindowState;
            MemHUD.Value = PBox.Height - 50;
            MemHUDV.Value = PBoxV.Height - 50;
            Redraw();
        }

        private void BtnSaveImg_Click(object sender, EventArgs e)
        {
            DlgSave.Filter = "Png Images|*.png";
            if (DlgSave.ShowDialog() == DialogResult.OK)
            {
                PictureBox box;
                int w, h;
                bool resize = false;
                
                // do we have to increae image size?
                if (MemTabs.SelectedIndex == 0)
                {
                    box = PBox;
                    if (ChkHorizontal.Checked)
                    {
                        w = Math.Max(PBox.Width, x + 30 + (int)MemHUD.Value);
                        h = PBox.Height;
                        resize = (w != PBox.Width);
                    }
                    else
                    {
                        h = Math.Max(PBox.Height, y + 30 + (int)MemHUD.Value);
                        w = PBox.Width;
                        resize = (h != PBox.Height);
                    }
                }
                else
                {
                    box = PBoxV;
                    if (ChkHorizontal.Checked)
                    {
                        w = Math.Max(PBoxV.Width, x + 30 + (int)MemHUDV.Value);
                        h = PBox.Height;
                        resize = (w != PBoxV.Width);
                    }
                    else
                    {
                        h = Math.Max(PBoxV.Height, y + 30 + (int)MemHUDV.Value);
                        w = PBoxV.Width;
                        resize = (h != PBoxV.Height);
                    }
                }

                if (resize)
                {
                    Image img = box.Image;
                    Bitmap b = new Bitmap(w, h);
                    box.Image = b;
                    Redraw();
                    box.Image = img;
                    using (Graphics g = Graphics.FromImage(b))
                    {
                        g.DrawImage(box.Image, 0, 0);
                        using (Brush br = new SolidBrush(Color.Black))
                        {
                            if (box == PBox)
                                g.DrawString("Physical", Font, br, 10, 10);
                            else
                                g.DrawString("Linear", Font, br, PBox.Width + 10, 10);
                        }
                    }
                    b.Save(DlgSave.FileName, System.Drawing.Imaging.ImageFormat.Png);
                    b.Dispose();
                }
                else
                {
                    using (Graphics g = Graphics.FromImage(box.Image))
                    {
                        using (Brush br = new SolidBrush(Color.Black))
                        {
                            if (box == PBox)
                                g.DrawString("Physical", Font, br, 10, 10);
                            else
                                g.DrawString("Linear", Font, br, 10, 10);
                        }
                    }
                    box.Image.Save(DlgSave.FileName, System.Drawing.Imaging.ImageFormat.Png);
                }
                Redraw();
            }
        }

        private void BtnLoadLayout_Click(object sender, EventArgs e)
        {
            // load layout
            DlgOpen.Filter = "Layouts|*.xml";
            if (DlgOpen.ShowDialog() == DialogResult.OK)
            {
                MemItem m;
                LstPhys.Items.Clear();
                LstVirtual.Items.Clear();
                XmlTextReader r = new XmlTextReader(DlgOpen.FileName);

                while (r.Read() && !r.EOF)
                {
                    if (r.Name == "Mem")
                    {
                        r.MoveToAttribute("SizeP");
                        TxtSize.Text = r.Value;
                        r.MoveToAttribute("SizeV");
                        TxtSizeV.Text = r.Value;
                        r.MoveToAttribute("FSize");
                        SizeUD.Value = decimal.Parse(r.Value);
                        if(r.MoveToAttribute("X"))
                            x = int.Parse(r.Value);
                        if(r.MoveToAttribute("XV"))
                            xV = int.Parse(r.Value);
                        if (r.MoveToAttribute("Horizontal"))
                            ChkHorizontal.Checked = (r.Value == "true");
                        else
                            ChkHorizontal.Checked = false;
                        if (r.MoveToAttribute("W"))
                        {
                            if (r.Value == "" || r.Value == "A")
                                MemHUD.Value = PBox.Height - 50;
                            else
                                MemHUD.Value = decimal.Parse(r.Value);
                        }
                        else
                        {
                            MemHUD.Value = PBox.Height - 50;
                        }
                        if (r.MoveToAttribute("WV"))
                        {
                            if (r.Value == "" || r.Value == "A")
                                MemHUDV.Value = PBoxV.Height - 50;
                            else
                                MemHUDV.Value = decimal.Parse(r.Value);
                        }
                        else
                        {
                            MemHUDV.Value = PBox.Height - 50;
                        }
                        _font = new Font(_font.FontFamily, (float)SizeUD.Value, FontStyle.Regular, GraphicsUnit.Pixel);
                        size = Parse(TxtSize.Text);
                        sizeV = Parse(TxtSizeV.Text);
                    }
                    else if (r.Name == "Physical")
                    {
                        if (r.NodeType == XmlNodeType.Element)
                        {
                            if (r.MoveToAttribute("Break"))
                            {
                                ChkBreak.Checked = (r.Value == "true");
                                r.MoveToAttribute("BFrom");
                                TxtBreakF.Text = r.Value;
                                r.MoveToAttribute("BTo");
                                TxtBreakT.Text = r.Value;
                            }
                        }
                        while (r.Read() && !r.EOF && !(r.NodeType == XmlNodeType.EndElement && r.Name == "Physical"))
                        {
                            if (r.Name == "Item")
                            {
                                m = new MemItem();
                                r.MoveToAttribute("Name");
                                m.Name = r.Value;
                                r.MoveToAttribute("Color");
                                m.Color = System.Drawing.ColorTranslator.FromHtml(r.Value);
                                r.MoveToAttribute("From");
                                m.From = Parse(r.Value);
                                r.MoveToAttribute("To");
                                m.To = Parse(r.Value);
                                InsertSorted(LstPhys,m);
                            }
                        }
                    }
                    else if (r.Name == "Virtual")
                    {
                        if (r.NodeType == XmlNodeType.Element)
                        {
                            if (r.MoveToAttribute("Break"))
                            {
                                ChkBreakV.Checked = (r.Value == "true");
                                r.MoveToAttribute("BFrom");
                                TxtBreakFV.Text = r.Value;
                                r.MoveToAttribute("BTo");
                                TxtBreakTV.Text = r.Value;
                            }
                        }
                        while (r.Read() && !r.EOF && !(r.NodeType == XmlNodeType.EndElement && r.Name == "Physical"))
                        {
                            if (r.Name == "Item")
                            {
                                m = new MemItem();
                                r.MoveToAttribute("Name");
                                m.Name = r.Value;
                                r.MoveToAttribute("Color");
                                m.Color = System.Drawing.ColorTranslator.FromHtml(r.Value);
                                r.MoveToAttribute("From");
                                m.From = Parse(r.Value);
                                r.MoveToAttribute("To");
                                m.To = Parse(r.Value);
                                InsertSorted(LstVirtual,m);
                            }
                        }
                    }
                }
                r.Close();
                _fileName = DlgOpen.FileName;
                Text = "MemGraph 0.3 - " + DlgOpen.FileName.Substring(DlgOpen.FileName.LastIndexOf("\\")+1);
                BtnAdd.Visible = true;
                BtnSave.Visible = false;
                edit = null;
                BtnSaveLayout.Enabled = true;
                if (ChkHAuto.Checked)
                    MemHUD.Value = (ChkHorizontal.Checked)? PBox.Width - 50 : PBox.Height - 50;
                if (ChkHAutoV.Checked)
                    MemHUDV.Value = (ChkHorizontal.Checked) ? PBoxV.Width - 50 : PBoxV.Height - 50;
                Redraw();
            }
        }

        private void BtnSaveLayout_Click(object sender, EventArgs e)
        {
            // save layout
            Save(_fileName);
        }

        private void BtnClearAll_Click(object sender, EventArgs e)
        {
            ListBox l = (MemTabs.SelectedIndex == 0) ? LstPhys : LstVirtual; 
            l.Items.Clear();
            Redraw();
        }

        private ulong Parse(string s)
        {
            if (s.StartsWith("0x"))
                s = s.Substring(2);
            return ulong.Parse(s, System.Globalization.NumberStyles.HexNumber);
        }

        private string IntToHex(ulong x)
        {
            return String.Format("0x{0:X}", x);
        }

        private void SaveAs_Click(object sender, EventArgs e)
        {
            Save(null);
        }

        string _fileName = null;

        private void Save(string fileName)
        {
            DlgSave.Filter = "Layouts|*.xml";
            if (fileName == null)
            {
                if (DlgSave.ShowDialog() != DialogResult.OK)
                    return;
                fileName = DlgSave.FileName;
            }
            XmlTextWriter w = new XmlTextWriter(fileName, null);
            w.WriteStartElement("Mem");
            w.WriteAttributeString("SizeP", TxtSize.Text);
            w.WriteAttributeString("SizeV", TxtSizeV.Text);
            w.WriteAttributeString("FSize", SizeUD.Value.ToString());
            w.WriteAttributeString("X", x.ToString());
            w.WriteAttributeString("XV", xV.ToString());
            w.WriteAttributeString("Horizontal", (ChkHorizontal.Checked)? "true" : "false");
            if(ChkHAuto.Checked)
                w.WriteAttributeString("W", "A");
            else
                w.WriteAttributeString("W", MemHUD.Value.ToString());
            if (ChkHAutoV.Checked)
                w.WriteAttributeString("WV", "A");
            else
                w.WriteAttributeString("WV", MemHUDV.Value.ToString());
            w.WriteStartElement("Physical");
            w.WriteAttributeString("Break", (ChkBreak.Checked) ? "true" : "false");
            w.WriteAttributeString("BFrom", TxtBreakF.Text);
            w.WriteAttributeString("BTo", TxtBreakT.Text);
            foreach (MemItem it in LstPhys.Items)
            {
                w.WriteStartElement("Item");
                w.WriteAttributeString("From", IntToHex(it.From));
                w.WriteAttributeString("To", IntToHex(it.To));
                w.WriteAttributeString("Name", it.Name);
                w.WriteAttributeString("Color", System.Drawing.ColorTranslator.ToHtml(it.Color));
                w.WriteEndElement();
            }
            w.WriteEndElement();
            w.WriteStartElement("Virtual");
            w.WriteAttributeString("Break", (ChkBreakV.Checked) ? "true" : "false");
            w.WriteAttributeString("BFrom", TxtBreakFV.Text);
            w.WriteAttributeString("BTo", TxtBreakTV.Text);
            foreach (MemItem it in LstVirtual.Items)
            {
                w.WriteStartElement("Item");
                w.WriteAttributeString("From", IntToHex(it.From));
                w.WriteAttributeString("To", IntToHex(it.To));
                w.WriteAttributeString("Name", it.Name);
                w.WriteAttributeString("Color", System.Drawing.ColorTranslator.ToHtml(it.Color));
                w.WriteEndElement();
            }
            w.WriteEndElement();
            w.WriteEndElement();
            w.Close();
            Text = "MemGraph 0.3 - " + DlgOpen.FileName.Substring(DlgOpen.FileName.LastIndexOf("\\") + 1);
        }

        private void BtnSave_Click(object sender, EventArgs e)
        {
            edit.Name = TxtName.Text;
            edit.From = Parse(TxtFrom.Text);
            edit.To = Parse(TxtTo.Text);
            edit.Color = LColor.BackColor;
            
            TxtName.Text = "";
            TxtFrom.Text = "";
            TxtTo.Text = "";
            Redraw();

            BtnAdd.Visible = true;
            BtnSave.Visible = false;

            InsertSorted(edit.List, edit);

            edit = null;
        }

        private void MemTabs_SelectedIndexChanged(object sender, EventArgs e)
        {
            LstPhys.Visible = (MemTabs.SelectedIndex == 0);
            LblLengthP.Visible = (MemTabs.SelectedIndex == 0);
            TxtSize.Visible = (MemTabs.SelectedIndex == 0);
            LstVirtual.Visible = (MemTabs.SelectedIndex == 1);
            LblLengthL.Visible = (MemTabs.SelectedIndex == 1);
            TxtSizeV.Visible = (MemTabs.SelectedIndex == 1);
            MemHUDV.Visible = (MemTabs.SelectedIndex == 1);
            ChkHAutoV.Visible = (MemTabs.SelectedIndex == 1);
            MemHUD.Visible = (MemTabs.SelectedIndex == 0);
            ChkHAuto.Visible = (MemTabs.SelectedIndex == 0);
            Grp.Text = (MemTabs.SelectedIndex == 0) ? "Physical" : "Linear";
            ChkBreak.Visible = (MemTabs.SelectedIndex == 0);
            TxtBreakF.Visible = (MemTabs.SelectedIndex == 0);
            TxtBreakT.Visible = (MemTabs.SelectedIndex == 0);
            ChkBreakV.Visible = (MemTabs.SelectedIndex == 1);
            TxtBreakFV.Visible = (MemTabs.SelectedIndex == 1);
            TxtBreakTV.Visible = (MemTabs.SelectedIndex == 1);
            if (MemTabs.SelectedIndex == 0)
                PBox.Invalidate();
            else
                PBoxV.Invalidate();
        }

        private void Lst_DrawItem(object sender, DrawItemEventArgs e)
        {
            e.DrawBackground();
            
            Brush myBrush = Brushes.Black;

            MemItem itm = (MemItem)((ListBox)sender).Items[e.Index];
            string txt = itm.ToString();

            SizeF s = e.Graphics.MeasureString(txt, e.Font);

            using (Brush b = new SolidBrush(itm.Color))
                e.Graphics.FillRectangle(b, 2, e.Bounds.Top, e.Bounds.Height, e.Bounds.Height);

            Rectangle r = e.Bounds;
            StringFormat f = new StringFormat(StringFormat.GenericDefault);
            f.LineAlignment = StringAlignment.Center;
            r.Offset(r.Height + 7, 0);
            e.Graphics.DrawString(txt, e.Font, myBrush, r, f);
            e.DrawFocusRectangle();
        }

        private void SizeUD_ValueChanged(object sender, EventArgs e)
        {
            _font = new Font(_font.FontFamily, (float)SizeUD.Value, FontStyle.Regular, GraphicsUnit.Pixel);
            Redraw();
        }

        private void MemGraph_ResizeEnd(object sender, EventArgs e)
        {
            PBox.Width = PBox.Width + (Width - fsizeX);
            PBox.Height = PBox.Height + (Height - fsizeY);
            PBoxV.Width = PBoxV.Width + (Width - fsizeX);
            PBoxV.Height = PBoxV.Height + (Height - fsizeY);
            PBox.Image = new Bitmap(PBox.Width, PBox.Height);
            PBoxV.Image = new Bitmap(PBoxV.Width, PBoxV.Height);
            if (ChkHAuto.Checked)
                MemHUD.Value = (ChkHorizontal.Checked) ? PBox.Width - 50 : PBox.Height - 50;
            if (ChkHAutoV.Checked)
                MemHUDV.Value = (ChkHorizontal.Checked) ? PBoxV.Width - 50 : PBoxV.Height - 50;
            Redraw();    
        }

        private void MemGraph_ResizeBegin(object sender, EventArgs e)
        {
            fsizeX = Width;
            fsizeY = Height;
        }
        
        private void PBox_MouseDown(object sender, MouseEventArgs e)
        {
            dragStart = e.X;
            xStart = x;
            dragStartY = e.Y;
            yStart = y;
            dragging = true;
        }

        private void PBox_MouseUp(object sender, MouseEventArgs e)
        {
            dragging = false;
        }

        private void PBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (dragging)
            {
                x = xStart + (e.X - dragStart);
                y = yStart + (e.Y - dragStartY);
            }
            Redraw();
        }

        private void PBoxV_MouseDown(object sender, MouseEventArgs e)
        {
            dragStart = e.X;
            xStart = xV;
            dragStartY = e.Y;
            yStart = yV;
            dragging = true;
        }

        private void PBoxV_MouseUp(object sender, MouseEventArgs e)
        {
            dragging = false;
        }

        private void PBoxV_MouseMove(object sender, MouseEventArgs e)
        {
            if (dragging)
            {
                xV = xStart + (e.X - dragStart);
                yV = yStart + (e.Y - dragStartY);
            }
            Redraw();
        }

        private void LstVirtual_DoubleClick(object sender, EventArgs e)
        {
            if (LstVirtual.SelectedItem != null)
            {
                edit = (MemItem)LstVirtual.SelectedItem;
                TxtName.Text = edit.Name;
                TxtFrom.Text = IntToHex(edit.From);
                TxtTo.Text = IntToHex(edit.To);
                LColor.BackColor = edit.Color;
                BtnAdd.Visible = false;
                BtnSave.Visible = true;
            }
        }

        private void LstPhys_DoubleClick(object sender, EventArgs e)
        {
            if (LstPhys.SelectedItem != null)
            {
                edit = (MemItem)LstPhys.SelectedItem;
                TxtName.Text = edit.Name;
                TxtFrom.Text = IntToHex(edit.From);
                TxtTo.Text = IntToHex(edit.To);
                LColor.BackColor = edit.Color;
                BtnAdd.Visible = false;
                BtnSave.Visible = true;
            }
        }

        private void MemGraph_SizeChanged(object sender, EventArgs e)
        {
            if (WindowState != OldState)
            {
                if (WindowState != FormWindowState.Minimized)
                {
                    PBox.Width = MemTabs.SelectedTab.Width;
                    PBox.Height = MemTabs.SelectedTab.Height;
                    PBoxV.Width = MemTabs.SelectedTab.Width;
                    PBoxV.Height = MemTabs.SelectedTab.Height;
                    PBox.Image = new Bitmap(PBox.Width, PBox.Height);
                    PBoxV.Image = new Bitmap(PBoxV.Width, PBoxV.Height);
                    if (ChkHAuto.Checked)
                        MemHUD.Value = (ChkHorizontal.Checked) ? PBox.Width - 50 : PBox.Height - 50;
                    if (ChkHAutoV.Checked)
                        MemHUDV.Value = (ChkHorizontal.Checked) ? PBoxV.Width - 50 : PBoxV.Height - 50;
                }
                OldState = WindowState;
                if (WindowState != FormWindowState.Minimized) Redraw(); 
            }
        }

        private void InsertSorted(ListBox l, MemItem itm)
        {
            if (l.Items.Contains(itm)) l.Items.Remove(itm);
            int i = 0;
            while (i < l.Items.Count)
            {
                MemItem it = (MemItem)l.Items[i];
                if(itm.From < it.From)
                {
                    l.Items.Insert(i, itm);
                    itm.List = l;
                    return;
                }
                i++;
            }
            itm.List = l;
            l.Items.Insert(l.Items.Count, itm);
        }

        private void ChkHAuto_CheckedChanged(object sender, EventArgs e)
        {
            MemHUD.Enabled = !ChkHAuto.Checked;
            MemHUDV.Enabled = !ChkHAutoV.Checked;
        }

        private void MemHUDV_ValueChanged(object sender, EventArgs e)
        {
            Redraw();
        }

        private void MemHUD_ValueChanged(object sender, EventArgs e)
        {
            Redraw();
        }

        #region Vertical Layout
        private void RedrawV(PictureBox box, ListBox l, int x, int y, ulong fromMem, ulong toMem, int ht, bool clear)
        {
            int w = 80;
            int h = ht;
            using (Graphics g = Graphics.FromImage(box.Image))
            {
                if(clear) g.Clear(Color.LightGray);

                using (Brush b = new SolidBrush(Color.White))
                {
                    g.FillRectangle(b, x, y, w, h);
                }

                int i = 0;
                while (i < l.Items.Count)
                {
                    MemItem it = (MemItem)l.Items[i];
                    
                    if (it.From > toMem) break; // finished
                    if (it.To < fromMem)
                    {
                        i++;
                        continue;
                    }

                    ulong oldTo, oldFrom;
                    bool printTo = true, printFrom = true;

                    oldTo = it.To;
                    if (it.To > toMem)
                    {
                        it.To = toMem;
                        printTo = false;
                    }
                    oldFrom = it.From;
                    if (it.From < fromMem)
                    {
                        it.From = fromMem;
                        printFrom = false;
                    }

                    int itmPos = it.itmPos = y + (int)(((Decimal)(it.From - fromMem) / (toMem - fromMem)) * h);
                    int itmPosTo = it.itmPosTo = y + (int)(((Decimal)(it.To - fromMem) / (toMem - fromMem)) * h);
                    using (Brush br = new SolidBrush(it.Color))
                    {
                        // calculate From X value and draw it
                        SizeF s;
                        if (printFrom)
                        {
                            s = g.MeasureString(IntToHex(it.From), _font);
                            it.yf = itmPos - (int)s.Height;
                            it.xf = CalcX(l, i, itmPos - (int)s.Height, x, y, h, s, g, true, (toMem - fromMem), fromMem);
                            g.DrawString(IntToHex(it.From), _font, br, it.xf, it.yf);
                        }
                        // draw to
                        if (printTo)
                        {
                            s = g.MeasureString(IntToHex(it.To), _font);
                            it.yt = itmPosTo - (int)s.Height;
                            it.xt = CalcX(l, i, itmPosTo - (int)s.Height, x, y, h, s, g, false, (toMem - fromMem), fromMem);
                            g.DrawString(IntToHex(it.To), _font, br, it.xt, it.yt);
                        }
                        // draw name
                        s = g.MeasureString(it.Name, _font);
                        it.yn = (int)(itmPos + ((itmPosTo - itmPos) / 2) - (s.Height / 2));
                        it.xn = CalcXName(l, i, it.yn, x, w, s, g, fromMem);
                        g.DrawString(it.Name, _font, br, it.xn, it.yn);
                        
                        // fill color
                        using (Pen p = new Pen(br))
                        {
                            p.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;
                            if (printFrom) g.DrawLine(p, it.xf, itmPos, x, itmPos); // from
                            if (printTo) g.DrawLine(p, it.xt, itmPosTo, x, itmPosTo); // to
                        }
                        g.FillRectangle(br, x, itmPos, w, itmPosTo - itmPos);

                        it.To = oldTo;
                        it.From = oldFrom;
                    }
                    i++;
                }
            }
            box.Invalidate();
        }

        public int CalcX(ListBox l, int i, int iy, int x, int y, int h, SizeF s, Graphics g, bool from, ulong size, ulong fromMem)
        {
            int j = i;
            int xp = x - 5 - (int)s.Width;
            while (j >= 0)
            {
                MemItem itm = (MemItem)l.Items[j];
                
                if (itm.To < fromMem) return xp;

                SizeF s2 = g.MeasureString(IntToHex(itm.From), _font);
                if (itm.From > fromMem || (itm.From == fromMem && (i != j)))
                {
                    // compara contra el From
                    if ((i != j || !from) && iy < itm.yf + s2.Height && (xp + s.Width > itm.xf - 5))
                        xp = itm.xf - (int)s.Width - 5;
                }
                // compara contra el To
                s2 = g.MeasureString(IntToHex(itm.To), _font);
                if (!(i == j) && iy < itm.yt + s2.Height && (xp + s.Width > itm.xt - 5))
                    xp = itm.xt - (int)s.Width - 5;
                j--;
            }
            return xp;
        }

        public int CalcXName(ListBox l, int i, int iy, int x, int w, SizeF s, Graphics g, ulong fromMem)
        {
            int j = i - 1;
            int xp = x + w + 10;
            while (j >= 0)
            {
                MemItem itm = (MemItem)l.Items[j];
                if (itm.From < fromMem) return xp;
                SizeF s2 = g.MeasureString(itm.Name, _font);
                if (iy < itm.yn + s2.Height && (xp < itm.xn + s2.Width))
                    xp = itm.xn + 5 + (int)s2.Width;
                j--;
            }
            return xp;
        }

        #endregion

        #region Horizontal Redraw
        private void RedrawH(PictureBox box, ListBox l, int x, int y, ulong fromMem, ulong toMem, int wt, bool clear)
        {
            int w = wt;
            int h = 80;
            using (Graphics g = Graphics.FromImage(box.Image))
            {
                if (clear) g.Clear(Color.LightGray);

                using (Brush b = new SolidBrush(Color.White))
                {
                    g.FillRectangle(b, x, y, w, h);
                }

                int i = 0;
                while (i < l.Items.Count)
                {
                    MemItem it = (MemItem)l.Items[i];

                    if (it.From > toMem) break; // finished
                    if (it.To < fromMem)
                    {
                        i++;
                        continue;
                    }

                    ulong oldTo, oldFrom;
                    bool printTo = true, printFrom = true;

                    oldTo = it.To;
                    if (it.To > toMem)
                    {
                        it.To = toMem;
                        printTo = false;
                    }
                    oldFrom = it.From;
                    if (it.From < fromMem)
                    {
                        it.From = fromMem;
                        printFrom = false;
                    }

                    int itmPos = it.itmPos = x + (int)(((Decimal)(it.From - fromMem) / (toMem - fromMem)) * w);
                    int itmPosTo = it.itmPosTo = x + (int)(((Decimal)(it.To - fromMem) / (toMem - fromMem)) * w);
                    using (Brush br = new SolidBrush(it.Color))
                    {
                        // draw From
                        SizeF s;
                        if (printFrom)
                        {
                            s = g.MeasureString(IntToHex(it.From), _font);
                            it.xf = itmPos - (int)s.Width;
                            it.yf = CalcY(l, i, it.xf, x, y, h, s, g, true, (toMem - fromMem));
                            g.DrawString(IntToHex(it.From), _font, br, it.xf, it.yf);
                        }
                        // draw to
                        if (printTo)
                        {
                            s = g.MeasureString(IntToHex(it.To), _font);
                            it.xt = itmPosTo - (int)s.Width;
                            it.yt = CalcY(l, i, it.xt, x, y, h, s, g, false, (toMem - fromMem));
                            g.DrawString(IntToHex(it.To), _font, br, it.xt, it.yt);
                        }
                        // draw name
                        s = g.MeasureString(it.Name, _font);
                        it.xn = (int)(itmPos + ((itmPosTo - itmPos) / 2) - s.Width);
                        it.yn = CalcYName(l, i, it.xn, y, h, s, g, (oldFrom < fromMem));
                        g.DrawString(it.Name, _font, br, it.xn, it.yn);
                        // fill color

                        using (Pen p = new Pen(br))
                        {
                            p.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;
                            if (printFrom) g.DrawLine(p, itmPos, it.yf, itmPos, y + h); // from
                            if (printTo) g.DrawLine(p, itmPosTo, it.yt, itmPosTo, y + h); // to
                            g.DrawLine(p, it.xn + s.Width, it.yn + s.Height, it.xn + s.Width, y); // name
                        }
                        g.FillRectangle(br, itmPos, y, itmPosTo - itmPos, h);
                        it.To = oldTo;
                        it.From = oldFrom;
                    }
                    i++;
                }
            }
            box.Invalidate();
        }

        public int CalcY(ListBox l, int i, int ix, int x, int y, int h, SizeF s, Graphics g, bool from, ulong size)
        {
            //CalcY(LstPhys, i, itemPos - Width, x, y, h, s, g, true, size); 

            int j = i;
            int yp = y + h + 20;
            while (j >= 0)
            {
                MemItem itm = (MemItem)l.Items[j];

                SizeF s2 = g.MeasureString(IntToHex(itm.From), _font);
                // compara contra el From
                if ((i != j || !from) && ix < itm.xf + s2.Width && (yp < itm.yf + s2.Height + 5))
                    yp = itm.yf + (int)s2.Height + 5;
                // compara contra el To
                s2 = g.MeasureString(IntToHex(itm.To), _font);
                if (!(i == j) && ix < itm.xt + s2.Width && (yp < itm.yt + s2.Height + 5))
                    yp = itm.yt + (int)s2.Height + 5;
                j--;
            }
            return yp;
        }

        public int CalcYName(ListBox l, int i, int ix, int y, int h, SizeF s, Graphics g, bool onBreak)
        {
            int j;
            if (onBreak)
                j = i;
            else
                j = i - 1;
            int yp = y - (int)s.Height - 10;
            while (j >= 0)
            {
                MemItem itm = (MemItem)l.Items[j];
                SizeF s2 = g.MeasureString(itm.Name, _font);
                if (ix < itm.xn + s2.Width && (yp > itm.yn - s2.Height - 5))
                    yp = itm.yn - (int)s2.Height - 5;
                j--;
            }
            return yp;
        }

        #endregion

        private void Redraw()
        {
            if (ChkHorizontal.Checked)
            {
                ulong f = Parse(TxtBreakF.Text);
                ulong t = Parse(TxtBreakT.Text);
                ulong fv = Parse(TxtBreakFV.Text);
                ulong tv = Parse(TxtBreakTV.Text);
                if (ChkBreak.Checked && t != f)
                {
                    int w1 = (int)((int)MemHUD.Value * (f / (decimal)(size - (t - f))));
                    RedrawH(PBox, LstPhys, x, y, 0, f, w1, true);
                    RedrawH(PBox, LstPhys, x + 16 + w1, y, t, size, (int)MemHUD.Value - w1, false);
                }
                else
                {
                    RedrawH(PBox, LstPhys, x, y, 0, size, (int)MemHUD.Value, true);
                }

                if (ChkBreakV.Checked && tv != fv)
                {
                    int w1V = (int)((int)MemHUDV.Value * (fv / (decimal)(sizeV - (tv - fv))));
                    RedrawH(PBoxV, LstVirtual, xV, yV, 0, fv, w1V, true);
                    RedrawH(PBoxV, LstVirtual, xV + 16 + w1V, yV, tv, sizeV, (int)MemHUDV.Value - w1V, false);
                }
                else
                {
                    RedrawH(PBoxV, LstVirtual, xV, yV, 0, sizeV, (int)MemHUDV.Value, true);
                }
            }
            else
            {
                ulong f = Parse(TxtBreakF.Text);
                ulong t = Parse(TxtBreakT.Text);
                ulong fv = Parse(TxtBreakFV.Text);
                ulong tv = Parse(TxtBreakTV.Text);
                if (ChkBreak.Checked && t != f)
                {
                    int h1 = (int)((int)MemHUD.Value * (f / (decimal)(size - (t - f))));
                    RedrawV(PBox, LstPhys, x, y, 0, f, h1, true);
                    RedrawV(PBox, LstPhys, x, y + 16 + h1, t, size, (int)MemHUD.Value - h1, false);
                }
                else
                {
                    RedrawV(PBox, LstPhys, x, y, 0, size, (int)MemHUD.Value, true);
                }

                if (ChkBreakV.Checked && tv != fv)
                {
                    int h1V = (int)((int)MemHUDV.Value * (fv / (decimal)(sizeV - (tv - fv))));
                    RedrawV(PBoxV, LstVirtual, xV, yV, 0, fv, h1V, true);
                    RedrawV(PBoxV, LstVirtual, xV, yV + 16 + h1V, tv, sizeV, (int)MemHUDV.Value - h1V, false);
                }
                else
                {
                    RedrawV(PBoxV, LstVirtual, xV, yV, 0, sizeV, (int)MemHUDV.Value, true);
                }
            }
        }

        private void ChkHorizontal_CheckedChanged(object sender, EventArgs e)
        {
            if (ChkHorizontal.Checked)
            {
                x = 30;
                y = 220;
                xV = 30;
                yV = 220;
                LblMH.Text = "Mem Width:";
            }
            else
            {
                x = 350;
                y = 30;
                xV = 350;
                yV = 30;
                LblMH.Text = "Mem Height:";
            }

            if (ChkHAuto.Checked)
                MemHUD.Value = (ChkHorizontal.Checked) ? PBox.Width - 50 : PBox.Height - 50;
            if (ChkHAutoV.Checked)
                MemHUDV.Value = (ChkHorizontal.Checked) ? PBoxV.Width - 50 : PBoxV.Height - 50;

            Redraw();
        }

        private void ChkHAutoV_CheckedChanged(object sender, EventArgs e)
        {
            MemHUD.Enabled = !ChkHAuto.Checked;
            MemHUDV.Enabled = !ChkHAutoV.Checked;
        }

        private void BtnGo_Click(object sender, EventArgs e)
        {
            ListBox l = ((MemTabs.SelectedIndex == 0) ? LstPhys : LstVirtual);
            MemItem it = (MemItem)l.SelectedItem;
            if (it != null)
            {
                // scroll to the item start
                if (ChkHorizontal.Checked)
                {
                    if (MemTabs.SelectedIndex == 0)
                    {
                        if (ChkBreak.Checked)
                        {
                            // calculate X considering the break
                            ulong f = Parse(TxtBreakF.Text);
                            ulong t = Parse(TxtBreakT.Text);
                            int w1 = (int)(MemHUD.Value * (f / (decimal)(size - (t - f))));
                            if (it.From <= f)
                                x = (int)(PBox.Width / 2) - (int)(((Decimal)it.From / f) * w1);
                            else
                                x = (int)(PBox.Width / 2) - (w1 + 16 + (int)(((Decimal)(it.From - t) / (decimal)(size - (t - f))) * (MemHUD.Value - w1)));
                        }
                        else
                        {
                            x = (int)(PBox.Width / 2) - (int)(((Decimal)it.From / size) * (int)MemHUD.Value);
                        }
                    }
                    else
                    {
                        if (ChkBreak.Checked)
                        {
                            // calculate X considering the break
                            ulong f = Parse(TxtBreakFV.Text);
                            ulong t = Parse(TxtBreakTV.Text);
                            int w1 = (int)(MemHUDV.Value * (f / (decimal)(sizeV - (t - f))));
                            if (it.From <= f)
                                xV = (int)(PBoxV.Width / 2) - (int)(((Decimal)it.From / f) * w1);
                            else
                                xV = (int)(PBoxV.Width / 2) - (w1 + 16 + (int)(((Decimal)(it.From - t) / (decimal)(sizeV - (t - f))) * (MemHUDV.Value - w1)));
                        }
                        else
                        {
                            xV = (int)(PBoxV.Width / 2) - (int)(((Decimal)it.From / sizeV) * (int)MemHUDV.Value);
                        }
                    }
                }
                else
                {
                    if (MemTabs.SelectedIndex == 0)
                    {
                        if (ChkBreak.Checked)
                        { 
                            // calculate X considering the break
                            ulong f = Parse(TxtBreakF.Text);
                            ulong t = Parse(TxtBreakT.Text);
                            int h1 = (int)(MemHUD.Value * (f / (decimal)(size - (t - f))));
                            if(it.From <= f)
                                y = (int)(PBox.Height / 2) - (int)(((Decimal)it.From / f) * h1);
                            else
                                y = (int)(PBox.Height / 2) - (h1 + 16 + (int)(((Decimal)(it.From - t) / (decimal)(size - (t - f))) * (MemHUD.Value - h1)));
                        }
                        else
                        {
                            y = (int)(PBox.Height / 2) - (int)(((Decimal)it.From / size) * (int)MemHUD.Value);
                        }
                    }
                    else
                    {
                        if (ChkBreak.Checked)
                        {
                            // calculate X considering the break
                            ulong f = Parse(TxtBreakFV.Text);
                            ulong t = Parse(TxtBreakTV.Text);
                            int h1 = (int)(MemHUDV.Value * (f / (decimal)(sizeV - (t - f))));
                            if (it.From <= f)
                                yV = (int)(PBoxV.Height / 2) - (int)(((Decimal)it.From / f) * h1);
                            else
                                yV = (int)(PBoxV.Height / 2) - (h1 + 16 + (int)(((Decimal)(it.From - t) / (decimal)(sizeV - (t - f))) * (MemHUDV.Value - h1)));
                        }
                        else
                        {
                            yV = (int)(PBoxV.Height / 2) - (int)(((Decimal)it.From / sizeV) * (int)MemHUDV.Value);
                        }
                    }
                }
                Redraw();
            }
        }

        private void ChkBreak_CheckedChanged(object sender, EventArgs e)
        {
            Redraw();
        }

        private void ChkBreakV_CheckedChanged(object sender, EventArgs e)
        {
            Redraw();
        }
    }

    public class MemItem
    {
        public string Name;
        public ulong From;
        public ulong To;
        public Color Color;

        public int xf;
        public int xt;
        public int xn;
        public int yn;
        public int itmPosTo;
        public int itmPos;
        public int yf;
        public int yt;

        public ListBox List;

        public override string ToString()
        {
            return Name + " (" + From.ToString("X") + " - " + To.ToString("X") +")";
        }
    }
}